;; Generated package description from dbus-codegen.el  -*- no-byte-compile: t -*-
(define-package "dbus-codegen" "0.1.0.20220306.62546" "Lisp code generation for D-Bus." '((cl-lib "0.5")) :commit "04114e4cbc8f422e6b77cfc202d1109c96316117" :url "https://elpa.gnu.org/packages/dbus-codegen.html" :authors '(("Daiki Ueno" . "ueno@gnu.org")) :maintainer '(nil . "emacs-devel@gnu.org") :keywords '("comm" "dbus" "convenience"))
